#define com_maroontress_libcgi_db_Tree_IMPORT
#define com_maroontress_libcgi_db_Box_IMPORT
#define com_maroontress_libcgi_db_XML_IMPORT
#define com_maroontress_libcgi_util_Tub_IMPORT

#include <sys/param.h>
#include <stdio.h>
#include <time.h>
#include <stdlib.h>
#include <com/maroontress/libcgi/db/Tree.h>
#include <com/maroontress/libcgi/db/XML.h>
#include <com/maroontress/libcgi/util/Tub.h>
#include <com/maroontress/libcgi/db/Box.h>
#include "BuildDetail.h"
#include "Serial.h"
#include "Project.h"
#include "Platform.h"

struct BuildDetail {
    struct Project *project;
    struct Serial *serial;
    char *xml;
};

void
BuildDetail_free(struct BuildDetail *d)
{
    if (d == NULL)
	return;
    Project_free(d->project);
    Serial_free(d->serial);
    free(d->xml);
    free(d);
}

static int
copyBuildDetail(struct Tub *tub, struct BuildDetail *d, struct Tree *builds)
{
    struct Box *box;
    int stamp, k, platforms;
    char date[32];
    time_t t;

    if ((box = Tree_getBox(builds, Serial_getKey(d->serial))) == NULL
        || Box_getInt(box, "stamp", &stamp) < 0
	|| Box_getInt(box, "platforms", &platforms) < 0) {
        return -1;
    }
    t = (time_t)stamp;
    strftime(date, sizeof(date), "%Y-%m-%d %T", localtime(&t));

    if (Tub_writeCString(tub,
			 "<status code=\"0\" message=\"OK\">\n"
			 "  <detail date=\"") < 0
        || XML_writeEscapedToTub(tub, date) < 0
	|| Tub_writeCString(tub, "\" revision=\"") < 0
	|| Box_writeXMLEscapedToTub(box, "revision", tub) < 0
	|| Tub_writeCString(tub, "\" author=\"") < 0
	|| Box_writeXMLEscapedToTub(box, "author", tub) < 0
	|| Tub_writeCString(tub, "\" status=\"") < 0
	|| Box_writeXMLEscapedToTub(box, "result", tub) < 0
	|| Tub_writeCString(tub, "\" />\n") < 0) {
	return -1;
    }
    for (k = 0; k < platforms; ++k) {
	/* Tub_print()snprintf()פ... */
	char code[16], statusString[16], *id;
	int status;

	snprintf(code, sizeof(code), "%d", k);
	if (Platform_getStringWithBox(box, k, "id", &id) < 0
	    || Platform_getIntWithBox(box, k, "status", &status) < 0) {
	    return -1;
	}
	snprintf(statusString, sizeof(statusString), "%d", status);
	if (Tub_writeCString(tub, "  <platform id=\"") < 0
	    || XML_writeEscapedToTub(tub, id) < 0
	    || Tub_writeCString(tub, "\" code=\"") < 0
	    || Tub_writeCString(tub, code) < 0
	    || Tub_writeCString(tub, "\" status=\"") < 0
	    || Tub_writeCString(tub, statusString) < 0
	    || Tub_writeCString(tub, "\" />\n") < 0) {
	    return -1;
	}
    }
    if (Tub_writeCString(tub, "</status>\n") < 0
        || Tub_writeByte(tub, 0) < 0) {
	return -1;
    }
    return 0;
}

static int
loadBuildDetailWithTree(struct BuildDetail *d, struct Tree *builds)
{
    struct Tub *tub;
    int s;

    if ((tub = Tub_create()) == NULL)
	return -1;
    s = (copyBuildDetail(tub, d, builds) < 0
	 || (d->xml = Tub_createByteArray(tub)) == NULL) ? -1 : 0;
    Tub_free(tub);
    return s;
}

static int
loadBuildDetail(struct BuildDetail *d)
{
    struct Tree *builds;
    int s;

    if ((builds = Tree_create(Project_getDBPath(d->project),
			      Tree_MODE_SHARED)) == NULL) {
        return -1;
    }
    s = loadBuildDetailWithTree(d, builds);
    Tree_free(builds);
    return s;
}

struct BuildDetail *
BuildDetail_create(const char *projectID, const char *serial)
{
    struct BuildDetail *d;

    if ((d = (struct BuildDetail *)malloc(sizeof(*d))) == NULL) {
	return NULL;
    }
    d->project = NULL;
    d->serial = NULL;
    d->xml = NULL;

    if ((d->project = Project_create(projectID)) == NULL
	|| (d->serial = Serial_create(atoi(serial))) == NULL
	|| loadBuildDetail(d) < 0) {
	BuildDetail_free(d);
        return NULL;
    }
    return d;
}

char *
BuildDetail_getResponse(struct BuildDetail *d)
{
    return d->xml;
}
