#define com_maroontress_libcgi_db_BoxKit_IMPORT
#define com_maroontress_libcgi_db_Box_IMPORT
#define com_maroontress_libcgi_db_Tree_IMPORT

#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <com/maroontress/libcgi/db/Box.h>
#include <com/maroontress/libcgi/db/BoxKit.h>
#include <com/maroontress/libcgi/db/Tree.h>
#include "BuildLog.h"
#include "Builder.h"
#include "Environment.h"
#include "Platform.h"
#include "PlatformList.h"
#include "Project.h"
#include "Serial.h"
#include "script.h"

#define FORMAT_BUILD_DB "projects/%s/build.db"
#define BIN_SH "/bin/sh"

struct Builder {
    /* ˷ */
    struct PlatformList *ls;
    char *projectID;
    int load;
    int mailEnabled;
    struct BuildLog *buildLog;
    struct Project *project;

    /* ꥢѹƤѤ */
    struct Serial *serial;
    char *revision;
    char logDir[MAXPATHLEN];
};

static int
updateSerial(struct Builder *b, int serial)
{
    Serial_free(b->serial);
    b->serial = NULL;
    free(b->revision);
    b->revision = NULL;

    if ((b->serial = Serial_create(serial)) == NULL) {
	return -1;
    }
    BuildLog_getBasePath(b->buildLog, b->serial, b->logDir, sizeof(b->logDir));
    return 0;
}

static int
initialize(struct Builder *b, struct BoxKit *kit)
{
    int serial;

    if ((b->projectID = strdup(BoxKit_getKey(kit))) == NULL
	|| (b->project = Project_create(b->projectID)) == NULL
	|| (b->buildLog = BuildLog_create(b->projectID)) == NULL) {
	return -1;
    }

    if (BoxKit_getInt(kit, "load", &b->load) < 0
	|| (b->load == 0 && BoxKit_setInt(kit, "load", 1) < 0)
        || BoxKit_getInt(kit, "mailEnabled", &b->mailEnabled) < 0
        || BoxKit_getInt(kit, "serial", &serial) < 0
        || BoxKit_setInt(kit, "serial", ++serial) < 0) {
	return -1;
    }
    return updateSerial(b, serial);
}

void
Builder_free(struct Builder *b)
{
    if (b == NULL)
	return;
    PlatformList_free(b->ls);
    free(b->projectID);
    BuildLog_free(b->buildLog);
    Project_free(b->project);
    Serial_free(b->serial);
    free(b->revision);
    free(b);
}

static int
addBuild(struct Builder *b, const char *revision, const char *author)
{
    struct Tree *builds;
    struct BoxKit *build;
    int s;

    if ((builds = Tree_create(Project_getDBPath(b->project),
			      Tree_MODE_EXCLUSIVE)) == NULL) {
	return -1;
    }
    s = ((build = Tree_getBoxKit(builds, Serial_getKey(b->serial))) == NULL
	 || BoxKit_setString(build, "revision", revision) < 0
	 || BoxKit_setString(build, "author", author) < 0
	 || BoxKit_setInt(build, "stamp", (int)time(NULL)) < 0
	 || BoxKit_setInt(build, "result", 0) < 0
	 || BoxKit_setInt(build, "platforms", 0) < 0
	 || BoxKit_close(build) < 0) ? -1 : 0;
    Tree_free(builds);
    return s;
}

struct Builder *
Builder_create(struct BoxKit *kit, const char *revision, const char *author)
{
    struct Builder *b;

    if ((b = (struct Builder *)malloc(sizeof(*b))) == NULL) {
	return NULL;
    }
    b->ls = NULL;
    b->projectID = NULL;
    b->serial = NULL;
    b->revision = NULL;
    b->buildLog = NULL;

    if ((b->ls = PlatformList_create(kit)) == NULL
	|| initialize(b, kit) < 0
	|| addBuild(b, revision, author) < 0) {
	Builder_free(b);
	return NULL;
    }
    return b;
}

static int
setResult(struct Builder *b, int result)
{
    struct Tree *builds;
    struct BoxKit *kit;
    int s;

    if ((builds = Tree_create(Project_getDBPath(b->project),
			      Tree_MODE_EXCLUSIVE)) == NULL) {
	return -1;
    }
    s = ((kit = Tree_getBoxKit(builds, Serial_getKey(b->serial))) == NULL
	 || BoxKit_setInt(kit, "result", result) < 0
	 || BoxKit_close(kit) < 0) ? -1 : 0;
    Tree_free(builds);
    return s;
}

struct Serial *
Builder_getSerial(struct Builder *b)
{
    return b->serial;
}

char *
Builder_getProjectID(struct Builder *b)
{
    return b->projectID;
}

static char *
getRevision(struct Builder *b)
{
    struct Tree *builds;
    struct Box *box;
    char *revision, *key = Serial_getKey(b->serial);

    free(b->revision);
    b->revision = NULL;

    if ((builds = Tree_create(Project_getDBPath(b->project),
			      Tree_MODE_SHARED)) == NULL) {
	return NULL;
    }
    b->revision = (((box = Tree_getBox(builds, key)) == NULL
		    || Box_getString(box, "revision", &revision) < 0)
		   ? NULL : strdup(revision));
    Tree_free(builds);
    return b->revision;
}

static int
buildPlatform(struct Builder *b, struct Platform *p, int k)
{
    char logFile[MAXPATHLEN];

    BuildLog_getPath(b->buildLog, b->serial, k, logFile, sizeof(logFile));
    return script(logFile, p->host, p->command, b->revision);
}

static int
addPlatformStatus(struct Builder *b, struct Platform *p, int k, int status)
{
    struct Tree *builds;
    struct BoxKit *kit;
    int s, platforms;

    if ((builds = Tree_create(Project_getDBPath(b->project),
			      Tree_MODE_EXCLUSIVE)) == NULL)
	return -1;
    s = ((kit = Tree_getBoxKit(builds, Serial_getKey(b->serial))) == NULL
	 || Platform_setStringWithBoxKit(kit, k, "id", p->id) < 0
	 || Platform_setIntWithBoxKit(kit, k, "status", status) < 0
	 || Platform_setStringWithBoxKit(kit, k, "charset", p->charset) < 0
	 || BoxKit_getInt(kit, "platforms", &platforms) < 0
	 || BoxKit_setInt(kit, "platforms", platforms + 1) < 0
	 || BoxKit_close(kit) < 0) ? -1 : 0;
    Tree_free(builds);
    return s;
}

static int
compressLog(struct Builder *b, int k, struct Environment *e)
{
    static const char *const COMMAND[] = {BIN_SH, "post-build.sh", NULL};
    struct Serial *s = Builder_getSerial(b);
    char serial[16], platformCode[16];
    int pid, status;

    if ((pid = fork()) < 0)
	return -1;
    if (pid != 0) {
	if (waitpid(pid, &status, 0) < 0)
	    return -1;
	return (!WIFEXITED(status) || WEXITSTATUS(status) != 0) ? -1 : 0;
    }
    snprintf(serial, sizeof(serial), "%d", Serial_getInt(s));
    snprintf(platformCode, sizeof(platformCode), "%d", k);
    if (Environment_setRaw(e, "serial", serial) < 0
	|| Environment_setRaw(e, "platformCode", platformCode) < 0
	|| Environment_setRaw(e, "logDir", Builder_getLogDir(b)) < 0
	|| Environment_execute(e, COMMAND) < 0) {
	_exit(1);
    }
    _exit(0);
}

int
Builder_run(struct Builder *b, struct Environment *e)
{
    struct Platform *p;
    unsigned int k, n, succeeded;
    int status;

    if (getRevision(b) == NULL) {
	return -1;
    }
    if (mkdir(b->logDir, 0755) < 0 && errno != EEXIST) {
	return -1;
    }

    succeeded = 0;
    n = PlatformList_getPlatformSize(b->ls);
    for (k = 0; k < n; ++k) {
	p = PlatformList_getPlatform(b->ls, k);
	if ((status = buildPlatform(b, p, k)) == Builder_SUCCESS) {
	    ++succeeded;
	}
	if (compressLog(b, k, e) < 0) {
	    return -1;
	}
	if (addPlatformStatus(b, p, k, status) < 0) {
	    return -1;
	}
    }

    status = (succeeded == n) ? Builder_SUCCESS : Builder_FAIL;
    if (setResult(b, status) < 0) {
	return -1;
    }
    return status;
}

int
Builder_setSerial(struct Builder *b, int serial)
{
    return updateSerial(b, serial);
}

int
Builder_getLoad(struct Builder *b)
{
    return b->load;
}

int
Builder_getMailEnabled(struct Builder *b)
{
    return b->mailEnabled;
}

static int
getPrevResult(struct Builder *b, struct Tree *builds)
{
    int num, status;
    struct Box *build;

    if ((num = Serial_getInt(b->serial)) <= 1)
	return -1;
    updateSerial(b, --num);
    if ((build = Tree_getBox(builds, Serial_getKey(b->serial))) == NULL
	|| Box_getInt(build, "result", &status) < 0) {
	return -1;
    }
    return status;
}

int
Builder_findPrevious(struct Builder *b)
{
    struct Tree *builds;
    int status;

    if ((builds = Tree_create(Project_getDBPath(b->project),
			      Tree_MODE_SHARED)) == NULL) {
	return 0;
    }
    while ((status = getPrevResult(b, builds)) == Builder_FAIL)
	;
    Tree_free(builds);
    if (status < 0 || status == Builder_SUCCESS)
	return 0;
    return 1;
}

char *
Builder_getLastRevision(struct Builder *b)
{
    return b->revision;
}

char *
Builder_getLogDir(struct Builder *b)
{
    return b->logDir;
}
